/*
 *************************************************************************
 *                                                                       *
 * Copyright 2000 Compaq Computer Corporation                            *
 *                                                                       *
 * COMPAQ Registered in U.S. Patent and Trademark Office.                *
 *                                                                       *
 *************************************************************************
 * IMPORTANT: Carefully read the License Terms below before              *
 * proceeding.  By use of these materials you agree to these terms.      *
 * If you do not agree to these terms, you may not use this software or  *
 * the accompanying documentation.                                       *
 *************************************************************************
 * LICENSE TERMS                                                         *
 * 1. GRANT                                                              *
 * Compaq Computer Corporation ("COMPAQ") grants you the right to use    *
 * and modify the following source code (the "Software") for use on any  *
 * number of computers.                                                  *
 * 2. COPYRIGHT                                                          *
 * The Software is owned by COMPAQ and its suppliers and is protected by *
 * copyright laws and international treaties.  Your use of the Software  *
 * and associated documentation is subject to the applicable copyright   *
 * laws and the express rights and restrictions of these terms.          *
 * 3. RESTRICTIONS                                                       *
 * You may not sell, rent, lease, license or otherwise commercially      *
 * transfer the Software.  You may not remove any copyright, trademark,  *
 * or other proprietary notices from the Software or the associated      *
 * documentation.   You are responsible for compliance with all          *
 * applicable export or re-export control laws and regulations if you    *
 * export the Software. This license is governed by and is to be         *
 * construed under the laws of the State of Texas.                       *
 *                                                                       *
 * DISCLAIMER OF WARRANTY AND LIABILITY                                  *
 * Compaq shall not be liable for technical or editorial errors or       *
 * omissions contained herein. The information contained herein is       *
 * subject to change without notice.                                     *
 *                                                                       *
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND.       *
 * THE ENTIRE RISK ARISING OUT OF THE USE OF THIS SOFTWARE REMAINS WITH  *
 * RECIPIENT.  IN NO EVENT SHALL COMPAQ BE LIABLE FOR ANY DIRECT,        *
 * CONSEQUENTIAL, INCIDENTAL, SPECIAL, PUNITIVE OR OTHER DAMAGES         *
 * WHATSOEVER (INCLUDING WITHOUT LIMITATION DAMAGES FOR LOSS OF BUSINESS *
 * PROFITS, BUSINESS INTERRUPTION, OR LOSS OF BUSINESS INFORMATION),     *
 * EVEN IF COMPAQ HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES    *
 * AND WHETHER IN AN ACTION OF CONTRACT OR TORT INCLUDING NEGLIGENCE.    *
 *                                                                       *
 * If you have any questions concerning this license, please contact:    *
 * Compaq Computer Corporation, Software Business Practices, ZKO1-2/D22, *
 * 110 Spit Brook Road, Nashua, NH. 03062-2698.                          *
 *                                                                       *
 *************************************************************************
 */


/*
**  MODULE DESCRIPTION:
**
**      VMSpoll - poll for OpenVMS
**
**      Implements poll via the select mechanism.
**
**	I added a whole bunch of debug stuff to this routine when I was
**	trouble-shooting a select error. It makes the code a little more
**	difficult but I thought I should leave it here in case its ever
**	needed again. Its only actually built in the DEBUG version.
**
**	The CLOEXEC support we just got with the new CRTL (ECO 5+) fixed one
**	of the infamous "error from select()" cases, but there's still another
**	one present. I am now even more convinced that the cause of this error
**	return from select() is due to the fact that a child has inherited a
**	socket, and when the child exits, the waiting select call in the
**	parent gets this hiccup. It seems totally benign and so I'm going to
**	have poll return a successful status to its caller (it will simulate
**	a timeout with no fd's ready situation).
**
**	Event Flags
**	===========
**
**	Support has now been added for event flags (as well as the usual
**	sockets). Events flags are denoted by passing them as a negative
**	fd number. Once set, whatever events were specified will be reflected
**	back in revents. When used with the X SetAsync functionality you get
**	the ability to include X events in your poll operation.
**
**	Since it is not possible to synchronise on both sockets and event
**	flags, we have to implement a polling mechanism. We perform a select
**	on the sockets but timeout every 50ms (configurable) so that we can
**	check the state of the events flags.
**
**	Basic logic is as follows:
**
**	- Walk pfds putting sockets into fd sets and EFs into EFmask.
**	- If any EF already set then set TIMEOUT to 0, otherwise set TIMEOUT
**	  to user timeout value but with a maximum of VMS_POLL_TIMER.
**	- If we have fds to check, then select(TIMEOUT) otherwise
**	  SLEEP(TIMEOUT).
**	- Check EFs. If no EF are set, no fds are ready, and user timeout
**	  is not yet reached, then select (or sleep) again and repeat this
**	  step.
**	- Walk fd sets and EFs putting results into pfds.
**
**	Current Restrictions:
**
**	- Only regular event flags are supported (EFN 0-63).
**
**	Other Notes:
**
**	- A pollfd of 0 is interpreted as EFN 0 and not FD 0. This seems the
**	  logical way to go with zero since FD 0 is stdin and clearly that
**	  can't be a socket. If we ever enhance select to also allow file FD's
**	  then we may want to rethink this.
**
**	- The poll timer we use when checking for event flags in now a
**	  sliding timer. It begins at VMS_POLL_TIMER_MIN, and increases
**	  by VMS_POLL_TIMER_DELTA each poll operation. The timer is never
**	  allowed to exceed VMS_POLL_TIMER_MAX. As soon as any activity
**	  is seen, the timer resets to the start value. Default values are:
**
**	    VMS_POLL_TIMER_MIN		 10 millisecs
**	    VMS_POLL_TIMER_DELTA	  1 millisecs
**	    VMS_POLL_TIMER_MAX		200 millisecs
*/

#define VMS_POLL_WITH_EF 1

#define MIN(a,b) ( (a) < (b) ? (a) : (b) )
#define MAX(a,b) ( (a) > (b) ? (a) : (b) )

#define FD_COPY(dest,src) memcpy(dest,src,sizeof(*(src)))

#include "vms_jackets.h"
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#ifdef VMS_POLL_WITH_EF
#include <pthread.h>
#include <starlet.h>
#include <string.h>
#endif

/*
** These are the poll events that we support. 
**
** There is another event that can come back from select as an exception, 
** but that's to do with remote terminals and shouldn't affect us since 
** we don't use them.
*/
#define  READ_EVENTS ( POLLIN  )
#define WRITE_EVENTS ( POLLOUT )
#define EXCEP_EVENTS ( POLLPRI )

#define NANOS_PER_MICRO 1000
#define MICROS_PER_MILLI 1000
#define MICROS_PER_SECOND 1000000

#ifdef VMS_POLL_WITH_EF
int poll_timer = -1;
int poll_timer_min, poll_timer_delta, poll_timer_max;
#endif

#ifdef DEBUG
int counter=0;
#endif

/*
** We call fcntl, and since we're a jacket, we don't get the auto-mappings.
** Therefore we need to prototype this ourselves.
*/
int GENERIC_FCNTL(int fildes, int cmd, int arg);

int GENERIC_POLL(struct pollfd pfds[], nfds_t npfds, int timeout)
{
    long n;
    struct pollfd *pfd, *epfd;
    fd_set rd, wt, ex, rd_copy, wt_copy, ex_copy;
    struct timeval tv, *tvp = NULL;
    int maxfd = -1, keep_polling=0;
    char errorString[255];
#ifdef VMS_POLL_WITH_EF
    int time_remaining=timeout, timer;
    unsigned int EF[2],EFmask[2]={0,0};
#endif

#ifdef DEBUG
    int dpf = (getenv(VMS_PREFIX "DEBUG_POLL_FULL") != NULL);
    int dps = (getenv(VMS_PREFIX "DEBUG_POLL_SHORT") != NULL);
    if (dpf) printf("GENERIC_POLL Entered (%x)\n",&pfds);
    if (dpf) printf("  (%x) npfds=%d, timeout=%d\n",&pfds, npfds,timeout);
    if (dps) printf("<%d",npfds);
#endif

#ifdef VMS_POLL_WITH_EF
    /*
    ** If we haven't gotten the value of the timers yet, then do so now.
    */
    if (poll_timer == -1) {
	{   char *p = getenv(VMS_PREFIX "POLL_TIMER_MIN");
	    if (p)
		poll_timer_min = atoi(p);
	    else
		poll_timer_min = 10; /* default is 10ms */
	}
	{   char *p = getenv(VMS_PREFIX "POLL_TIMER_DELTA");
	    if (p)
		poll_timer_delta = atoi(p);
	    else
		poll_timer_delta = 1; /* default is 1ms */
	}
	{   char *p = getenv(VMS_PREFIX "POLL_TIMER_MAX");
	    if (p)
		poll_timer_max = atoi(p);
	    else
		poll_timer_max = 200; /* default is 200ms */
	}
	poll_timer = poll_timer_min;
    }
#endif

    /*
    ** Zero out all the fd sets.
    */
    FD_ZERO(&rd);
    FD_ZERO(&wt);
    FD_ZERO(&ex);

    /*
    ** Walk all the fds, setting the the appropriate FD_SETs as we go, and
    ** clearing the revents. Anything negative is an event flag to watch.
    */
    for (pfd = pfds, epfd = pfd + npfds; pfd < epfd; pfd++) {
	int osfd = pfd->fd;
        short events = pfd->events;

#ifdef DEBUG
	if (dpf) {
	    printf("  (%x) BEFORE fd=%d, events=%d ( ",&pfds,osfd,events);
	    if (events & READ_EVENTS) printf("READ ");
	    if (events & WRITE_EVENTS) printf("WRITE ");
	    if (events & EXCEP_EVENTS) printf("EXCEPT ");
	    printf(")\n");
	}
#endif

	if (osfd > 0) {
	    if (osfd > maxfd) {
		maxfd = osfd;
	    }

	    if (events & READ_EVENTS) 
		FD_SET(osfd, &rd);

	    if (events & WRITE_EVENTS)
		FD_SET(osfd, &wt);

	    if (events & EXCEP_EVENTS)
		FD_SET(osfd, &ex);
	}
	else {
#ifdef VMS_POLL_WITH_EF
	    int flag = -osfd;
	    if (flag < 32)
		EFmask[0] |= ( 1 << flag );
	    else
		if (flag < 64)
		    EFmask[1] |= ( 1 << (flag-32) );
		else /* fall out of ifdef */
#endif
	    {
#ifdef DEBUG
		printf("*** Invalid FD - %d\n",osfd);
#endif
		errno = EBADF;
		return -1;
	    }
	}
	/* Assume we won't find any */
	pfd->revents = 0;
    }

    do {

#ifdef VMS_POLL_WITH_EF 
	int microsecs, rv;

	/*
	** If there are event flags to watch, then we wait a maximum of
	** poll timer. We also only want to wait whatever portion is left
	** of the user's timeout value.
	**
	** If we are not watching event flags, then we can just let
	** select do all the work.
	*/
	if (EFmask[0]|EFmask[1]) {
	    /*
	    ** Yes, we have some event flags to monitor.
	    */
	    if (timeout == INFTIM) {
		timer = poll_timer;
	    }
	    else {
		timer = MIN(timeout,poll_timer);
		if (timer > time_remaining)
		    timer = time_remaining;
		time_remaining -= timer;
	    }
	    /*
	    ** Check the event flags. If any are already set, then we
	    ** don't want select to wait at all.
	    */
	    if (EFmask[0]) {
		sys$readef( 0,&EF[0]);
		EF[0] &= EFmask[0];
	    }
	    else
		EF[0]=0;
	    if (EFmask[1]) {
		sys$readef(32,&EF[1]);
		EF[1] &= EFmask[1];
	    }
	    else
		EF[1]=0;
	    if (EF[0]|EF[1]) {
		/* Some are already set, so force no timeout */
		tv.tv_sec = 0;
		tv.tv_usec = 0;
	    }
	    else {
		/* None are currently set, so use computed timeout */
		microsecs = timer * MICROS_PER_MILLI ;
		tv.tv_sec = microsecs / MICROS_PER_SECOND ;
		tv.tv_usec = microsecs % MICROS_PER_SECOND ;
	    }
	    /* We are certainly using a timer */
	    tvp = &tv;
	}
	else {
	    /*
	    ** No event flags to watch.
	    ** Set the timeout value from the user-supplied value.
	    ** It comes in in milliseconds. Watch out for no timeout!
	    */
	    if (timeout != INFTIM) {
		int microsecs = timeout * MICROS_PER_MILLI ;
		tv.tv_sec = microsecs / MICROS_PER_SECOND ;
		tv.tv_usec = microsecs % MICROS_PER_SECOND ;
		tvp = &tv;
	    }
	}

#ifdef DEBUG
	if (EFmask[0])
	    if (dpf) printf("  (%x) Before - EF-low are x%x/x%x\n",&pfds,
		EF[0],EFmask[0]);
	if (EFmask[1])
	    if (dpf) printf("  (%x) Before - EF-high are x%x/x%x\n",&pfds,
		EF[1],EFmask[1]);
#endif

#else
	/*
	** Set up the timeout value. It comes in in milliseconds.
	*/
	if (timeout != INFTIM) {
	    int microsecs = timeout * MICROS_PER_MILLI ;
	    tv.tv_sec = microsecs / MICROS_PER_SECOND ;
	    tv.tv_usec = microsecs % MICROS_PER_SECOND ;
	    tvp = &tv;
	}
#endif

	/*
	** And do the select call. On successful completion, select() returns
	** the total number of bits set in the bit masks. Otherwise, -1 is
	** returned, and errno is set to indicate the error. 
	*/
#ifdef DEBUG
#ifdef VMS_POLL_WITH_EF
	if (dpf) printf("  (%x) Doing the SELECT, time=%d/%d/%d\n",&pfds,timer,time_remaining,timeout);
#else
	if (dpf) printf("  (%x) Doing the SELECT, time=%d\n",&pfds,timeout);
#endif
#endif
	assert(FD_SETSIZE > maxfd+1);  

#ifdef VMS_POLL_WITH_EF
	/*
	** If we weren't given any sockets, then just sleep instead of
	** selecting. We can do a much more thread friendly sleep than
	** select can.
	*/
	if (maxfd == -1) {
	    struct timespec ts;
	    ts.tv_sec = tv.tv_sec;
	    ts.tv_nsec = tv.tv_usec * NANOS_PER_MICRO;
	    pthread_delay_np(&ts);
	    n=0;
	}
	else {
	    /*
	    ** If we will ever need to loop, then make a copy of the fs sets.
	    ** This will be quicker than having to set them all up again.
	    */
	    if (time_remaining != 0) {
		FD_COPY(&rd_copy,&rd);
		FD_COPY(&wt_copy,&wt);
		FD_COPY(&ex_copy,&ex);
	    }
	    n = select(maxfd + 1, &rd, &wt, &ex, tvp);
	}
#else
	n = select(maxfd + 1, &rd, &wt, &ex, tvp);
#endif

#ifdef DEBUG
	if (dpf) printf("  (%x) Back from SELECT, status is %d\n",&pfds,n);
	if (dpf)
	    if (n<0)
		printf("  (%x) errno=%d, vaxc$errno=%d\n",&pfds,errno,vaxc$errno);
#endif

#ifdef VMS_POLL_WITH_EF
	/*
	** Check EFs again since they could have changed. If no EF are set,
	** no fds are ready, and user timeout is not yet reached, then we
	** want to select again;
	*/
	if (EFmask[0]) {
	    rv = sys$readef(0,&EF[0]);
	    EF[0] &= EFmask[0];
#ifdef DEBUG
	    if (dpf) printf("  (%x) After - EF-low are x%x/x%x\n",
		&pfds,EF[0],EFmask[0]);
#endif
	} else
	    EF[0] = 0;
	if (EFmask[1]) {
	    rv = sys$readef(32,&EF[1]);
	    EF[1] &= EFmask[1];
#ifdef DEBUG
	    if (dpf) printf("  (%x) After - EF-high are x%x/x%x\n",
		&pfds,EF[1],EFmask[1]);
#endif
	} else
	    EF[1] = 0;
	if ( (EF[0] == 0) && (EF[1] == 0) && (n == 0) && (time_remaining != 0) ) {
	    keep_polling = 1;
	    FD_COPY(&rd,&rd_copy);
	    FD_COPY(&wt,&wt_copy);
	    FD_COPY(&ex,&ex_copy);
	    poll_timer += poll_timer_delta;
	    poll_timer = MIN(poll_timer, poll_timer_max);
	}
	else
	    keep_polling = 0;
#endif

    }
    while (keep_polling) ;

    /*
    ** Walk the pfds and deal with the sockets.
    */
    if (n > 0) {
	/*
	** Success. Now walk all the fds again, this time setting the output
	** events for all those that are ready.
	*/
	n = 0;
	for (pfd = pfds, epfd = pfd + npfds; pfd < epfd; pfd++) {
	    int osfd = pfd->fd;
	    short events = pfd->events;
	    short revents = 0;

	    if (osfd > 0) {
#ifdef DEBUG
		if (dpf) {
		    printf("  (%x) AFTER fd=%d, events=%d ( ",&pfds,osfd,events);
		    if ((events & READ_EVENTS) && FD_ISSET(osfd, &rd)) printf("READ ");
		    if ((events & WRITE_EVENTS) && FD_ISSET(osfd, &wt)) printf("WRITE ");
		    if ((events & EXCEP_EVENTS) && FD_ISSET(osfd, &ex)) printf("EXCEPT ");
		    printf(")\n");
		}
#endif
		if ((events & READ_EVENTS) && FD_ISSET(osfd, &rd)) 
		    revents |= (events & READ_EVENTS);

		if ((events & WRITE_EVENTS) && FD_ISSET(osfd, &wt))
		    revents |= (events & WRITE_EVENTS);

		if ((events & EXCEP_EVENTS) && FD_ISSET(osfd, &ex))
		    revents |= (events & EXCEP_EVENTS);

		pfd->revents = revents;
		if (revents)
		    n++;
	    }
	}
	assert(n > 0);
    }
    else if (n < 0) {
	if (errno == EBADF) {
	    /* Find the bad fds */
	    n = 0;
	    for (pfd = pfds, epfd = pfd + npfds; pfd < epfd; pfd++) {
		if (pfd->fd > 0) {
		    if (GENERIC_FCNTL(pfd->fd, F_GETFL, 0) == -1) {
#ifdef DEBUG
			if (dpf) printf("  (%x) bad fd is %d\n",&pfds,pfd->fd);
#endif
			pfd->revents = POLLNVAL;
			n++;
		    }
		}
	    }
	    assert(n > 0);
	} else
	if ((errno == 1) && (vaxc$errno == 2096)) {
	    /* This is the weird case we treat as a spurious timeout */
            GENERIC_TRACE( "INFO - GENERIC_POLL detected spurious select return",
					"", "", "" );
	    n=0;
	} else {
	    sprintf( errorString,
            "GENERIC_POLL: select() call failed, n=%d, errno=%d, vaxc$errno=%d\n",
             n, errno, vaxc$errno);
            GENERIC_TRACE( errorString, "", "", "" );

	}
    }


#ifdef VMS_POLL_WITH_EF
    /*
    ** If all is well, then rewalk the pfds and this time take care of
    ** any event flags pollfds.
    */
    if ( (EF[0]|EF[1]) && (n >= 0) ) {
	for (pfd = pfds, epfd = pfd + npfds; pfd < epfd; pfd++) {
	    int osfd = pfd->fd;
	    if (osfd <= 0) {
		unsigned int flag = -osfd, bit;
		/* If the EF is set, then replicate events in revents */
		if (flag < 32)
		    bit = EF[0] & (1<<(flag));
		else
		    bit = EF[1] & (1<<(flag-32));
		if (bit) {
		    pfd->revents = pfd->events;
		    n++;
		}
#ifdef DEBUG
		if (dpf) {
		    printf("  (%x) AFTER fd=%d, events=%d ( ",&pfds,osfd,pfd->revents);
		    if (pfd->revents & READ_EVENTS) printf("READ ");
		    if (pfd->revents & WRITE_EVENTS) printf("WRITE ");
		    if (pfd->revents & EXCEP_EVENTS) printf("EXCEPT ");
		    printf(")\n");
		}
#endif
	    }
	}
    }
#endif

#ifdef DEBUG
    if (dpf)
	if (n)
	    printf("  (%x) n=%d <<<<=====================\n",&pfds,n);
    if (dpf) printf("GENERIC_POLL Leaving (%x)\n",&pfds);
    if (dps) {
	printf(">");
	if (!(++counter%10)) printf("\n");
    }
#endif
#ifdef VMS_POLL_WITH_EF
    /*
    ** If something really happened, then reset the poll timer.
    */
    if (n>0)
	poll_timer = poll_timer_min;
#endif

    return n;
}

